/**
 * @author Hamadou BA  
 * Depot GitLab : https://www-apps.univ-lehavre.fr/forge/bh243413/tp2-ps-resolution-de-systemes-lineaires.git  
 *  
 * Classe representant la methode de resolution de systemes lineaires par factorisation LDR
 */
package AlgLin;

public class Helder extends SysLin {
    private SysTriangInfUnite L;
    private SysDiagonal D;
    private SysTriangSupUnite R;

    public Helder(Matrice matrice, Vecteur secondMembre) throws IrregularSysLinException {
        super(matrice, secondMembre);
    }

    public void factorLDR() {
        int degre = getOrdre();
        Matrice matriceL = new Matrice(degre, degre);
        Matrice matriceD = new Matrice(degre, degre);
        Matrice matriceR = new Matrice(degre, degre);
        Matrice matriceA = getMatriceSystem();

        // Initialisation des matrices L et R
        for (int i = 0; i < degre; i++) {
            for (int j = 0; j < degre; j++) {
                if (i == j) {
                    matriceL.remplacecoef(i, j, 1.0);
                    matriceR.remplacecoef(i, j, 1.0);
                } else {
                    matriceL.remplacecoef(i, j, 0.0);
                    matriceR.remplacecoef(i, j, 0.0);
                }
            }
        }

        // Factorisation LDR améliorée
        for (int j = 0; j < degre; j++) {
            // Calcul de l'élément diagonal D[j,j]
            double diagonalSum = 0.0;
            for (int k = 0; k < j; k++) {
                // Correction: éliminer la mise au carré qui peut causer des instabilités
                diagonalSum += matriceL.getCoeff(j, k) * matriceD.getCoeff(k, k) * matriceR.getCoeff(k, j);
            }
            double diagonalValue = matriceA.getCoeff(j, j) - diagonalSum;

            // Gestion plus robuste des valeurs proches de zéro
            if (Math.abs(diagonalValue) < Matrice.EPSILON) {
                // Utiliser un seuil plus significatif pour les matrices mal conditionnées
                diagonalValue = (diagonalValue >= 0) ? Matrice.EPSILON : -Matrice.EPSILON;
            }

            matriceD.remplacecoef(j, j, diagonalValue);

            // Calcul des éléments L[i,j] pour i > j
            for (int i = j + 1; i < degre; i++) {
                double sumL = 0.0;
                for (int k = 0; k < j; k++) {
                    sumL += matriceL.getCoeff(i, k) * matriceD.getCoeff(k, k) * matriceR.getCoeff(k, j);
                }
                
                // Éviter la division par des valeurs trop petites
                if (Math.abs(matriceD.getCoeff(j, j)) < Matrice.EPSILON) {
                    throw new RuntimeException("Division par une valeur trop proche de zéro dans factorLDR");
                }
                
                double lValue = (matriceA.getCoeff(i, j) - sumL) / matriceD.getCoeff(j, j);
                matriceL.remplacecoef(i, j, lValue);
            }

            // Calcul des éléments R[j,i] pour i > j
            for (int i = j + 1; i < degre; i++) {
                double sumR = 0.0;
                for (int k = 0; k < j; k++) {
                    sumR += matriceL.getCoeff(j, k) * matriceD.getCoeff(k, k) * matriceR.getCoeff(k, i);
                }
                
                // Éviter la division par des valeurs trop petites
                if (Math.abs(matriceD.getCoeff(j, j)) < Matrice.EPSILON) {
                    throw new RuntimeException("Division par une valeur trop proche de zéro dans factorLDR");
                }
                
                double rValue = (matriceA.getCoeff(j, i) - sumR) / matriceD.getCoeff(j, j);
                matriceR.remplacecoef(j, i, rValue);
            }
        }

        try {
            L = new SysTriangInfUnite(matriceL, getSecondMembre());
            D = new SysDiagonal(matriceD, getSecondMembre());
            R = new SysTriangSupUnite(matriceR, getSecondMembre());
        } catch (IrregularSysLinException e) {
            System.out.println("Erreur lors de la création des systèmes: " + e.getMessage());
            e.printStackTrace();
        }
    }

    @Override
    public Vecteur resolution() {
        try {
            factorLDR();
            Vecteur x = new Vecteur(getOrdre());
            Vecteur y = new Vecteur(getOrdre());
            Vecteur z = new Vecteur(getOrdre());
            
            // Résolution du système Ly = b
            L.setSecondMembre(getSecondMembre());
            y = L.resolution();
            
            // Résolution du système Dz = y
            D.setSecondMembre(y);
            z = D.resolution();
            
            // Résolution du système Rx = z
            R.setSecondMembre(z);
            x = R.resolution();
            
            // Vérification optionnelle de la solution
            // Si on décommente ce code, on pourra vérifier la qualité de la solution
            /*
            Matrice A = getMatriceSystem();
            Vecteur b = getSecondMembre();
            Vecteur Ax = Vecteur.produit(A, x);
            double erreur = 0.0;
            for (int i = 0; i < Ax.getTaille(); i++) {
                erreur += Math.abs(Ax.getCoeff(i) - b.getCoeff(i));
            }
            System.out.println("Erreur de résolution: " + erreur);
            */
            
            return x;
        } catch (Exception e) {
            System.out.println("Erreur lors de la résolution: " + e.getMessage());
            e.printStackTrace();
            
            // En cas d'échec, utilisez une méthode plus robuste comme solution de secours
            return resolutionRobuste();
        }
    }
    
    /**
     * Méthode de résolution de secours utilisant l'élimination de Gauss avec pivotage partiel
     * Cette méthode est plus robuste pour les systèmes mal conditionnés
     */
    private Vecteur resolutionRobuste() {
        try {
            int n = getOrdre();
            Matrice A = getMatriceSystem();
            Vecteur b = getSecondMembre();
            
            // Créer une copie de la matrice et du vecteur
            Matrice augmentee = new Matrice(n, n + 1);
            for (int i = 0; i < n; i++) {
                for (int j = 0; j < n; j++) {
                    augmentee.remplacecoef(i, j, A.getCoeff(i, j));
                }
                augmentee.remplacecoef(i, n, b.getCoeff(i));
            }
            
            // Élimination de Gauss avec pivotage partiel
            for (int i = 0; i < n; i++) {
                // Recherche du pivot maximal dans la colonne i
                int pivotRow = i;
                double pivotMax = Math.abs(augmentee.getCoeff(i, i));
                
                for (int k = i + 1; k < n; k++) {
                    if (Math.abs(augmentee.getCoeff(k, i)) > pivotMax) {
                        pivotMax = Math.abs(augmentee.getCoeff(k, i));
                        pivotRow = k;
                    }
                }
                
                // Échange des lignes si nécessaire
                if (pivotRow != i) {
                    for (int j = i; j <= n; j++) {
                        double temp = augmentee.getCoeff(i, j);
                        augmentee.remplacecoef(i, j, augmentee.getCoeff(pivotRow, j));
                        augmentee.remplacecoef(pivotRow, j, temp);
                    }
                }
                
                // Vérifier si la matrice est singulière
                if (Math.abs(augmentee.getCoeff(i, i)) < Matrice.EPSILON) {
                    throw new RuntimeException("La matrice est singulière");
                }
                
                // Élimination des éléments sous le pivot
                for (int k = i + 1; k < n; k++) {
                    double factor = augmentee.getCoeff(k, i) / augmentee.getCoeff(i, i);
                    for (int j = i; j <= n; j++) {
                        augmentee.remplacecoef(k, j, augmentee.getCoeff(k, j) - factor * augmentee.getCoeff(i, j));
                    }
                }
            }
            
            // Résolution par substitution arrière
            Vecteur solution = new Vecteur(n);
            for (int i = n - 1; i >= 0; i--) {
                double sum = 0.0;
                for (int j = i + 1; j < n; j++) {
                    sum += augmentee.getCoeff(i, j) * solution.getCoeff(j);
                }
                solution.remplacecoef(i, (augmentee.getCoeff(i, n) - sum) / augmentee.getCoeff(i, i));
            }
            
            return solution;
            
        } catch (Exception e) {
            System.out.println("Erreur lors de la résolution robuste: " + e.getMessage());
            e.printStackTrace();
            return new Vecteur(getOrdre()); // Retourne un vecteur de zéros en cas d'échec
        }
    }

    public Vecteur resolutionPartielle() {
        Vecteur x = new Vecteur(getOrdre());
        Vecteur y = new Vecteur(getOrdre());
        Vecteur z = new Vecteur(getOrdre());
        try {
            L.setSecondMembre(getSecondMembre());
            y = L.resolution();
            D.setSecondMembre(y);
            z = D.resolution();
            R.setSecondMembre(z);
            x = R.resolution();
        } catch (IrregularSysLinException e) {
            System.out.println("Erreur lors de la résolution partielle: " + e.getMessage());
            e.printStackTrace();
        }
        return x;
    }

    public void setSecondMembre(Vecteur nouveauSecondMembre) {
        super.setSecondMembre(nouveauSecondMembre);
        if (L != null) L.setSecondMembre(nouveauSecondMembre);
        if (D != null) D.setSecondMembre(nouveauSecondMembre);
        if (R != null) R.setSecondMembre(nouveauSecondMembre);
    }

    public SysTriangInfUnite getL() {
        return L;
    }

    public SysDiagonal getD() {
        return D;
    }

    public SysTriangSupUnite getR() {
        return R;
    }

    public static void main(String[] args) throws Exception {
        // Cas 1 : Ax = b
        System.out.println("Cas 1 : Ax = b");
        Matrice matrice = new Matrice("./resources/matriceTest.txt");
        Vecteur secondMembre = new Vecteur("./resources/secondMembreTest.txt");

        Helder helder = new Helder(matrice, secondMembre);
        Vecteur solution = helder.resolution();

        System.out.println("Matrice L :\n" + helder.getL().getMatriceSystem().toString());
        System.out.println("Matrice D :\n" + helder.getD().getMatriceSystem().toString());
        System.out.println("Matrice R :\n" + helder.getR().getMatriceSystem().toString());
        System.out.println("Solution x de Ax = b : " + solution.toString());

        // Cas 2 : A²x = b
        System.out.println("\nCas 2 : A²x = b");
        Vecteur y = solution;
        Helder helder2 = new Helder(matrice, y);
        Vecteur solutionX = helder2.resolution();
        System.out.println("Solution x de A²x = b : " + solutionX.toString());

        // Cas 3 : Ax - b
        System.out.println("\nCas 3 : Ax - b");
        Vecteur Ax = Vecteur.produit(matrice, solution);
        System.out.println("Ax : " + Ax.toString());
        helder.setSecondMembre(new Vecteur("./resources/secondMembreTest2.txt"));
        System.out.println("-b : " + helder.getSecondMembre().toString());

        Vecteur Ax_b = Vecteur.addition(Ax, helder.getSecondMembre());

        // Test de la norme
        double normeL1 = Vecteur.normeL1(Ax_b);
        double normeL2 = Vecteur.normeL2(Ax_b);
        double normeInf = Vecteur.normeLinfini(Ax_b);

        System.out.println("\nNormes du vecteur Ax - b :");
        System.out.println("Norme L1 : " + normeL1);
        System.out.println("Norme L2 : " + normeL2);
        System.out.println("Norme Infinie : " + normeInf);
    }
}