package tp6.ps;

import AlgLin.Helder;
import AlgLin.IrregularSysLinException;
import AlgLin.Matrice;
import AlgLin.Vecteur;

/**
 * Classe représentant un modèle polynomiale pour l'approximation par moindres carrés
 * Le modèle est de la forme: p(x) = a0 + a1*x + a2*x^2 + ... + am*x^m
 */
public class ModPoly {
    private int degre;         // Degré du polynôme
    private Vecteur coefficients; // Coefficients ai du polynôme

    /**
     * Constructeur pour un modèle polynomiale de degré donné
     * @param degre Le degré du polynôme
     */
    public ModPoly(int degre) {
        this.degre = degre;
        this.coefficients = new Vecteur(degre + 1); // +1 car on a les coefficients de 0 à degre
    }

    /**
     * Calcule la valeur du polynôme en un point x
     * @param x Le point où évaluer le polynôme
     * @return La valeur p(x)
     */
    public double evaluer(double x) {
        double resultat = 0.0;
        for (int i = 0; i <= degre; i++) {
            resultat += coefficients.getCoeff(i) * Math.pow(x, i);
        }
        return resultat;
    }

    /**
     * Retourne la fonction de base i
     * @param i L'indice de la fonction de base
     * @param x Le point où évaluer la fonction
     * @return La valeur f_i(x) = x^i
     */
    private double fonctionDeBase(int i, double x) {
        return Math.pow(x, i);
    }

    /**
     * Identifie les coefficients du modèle polynomial selon le critère des moindres carrés
     * en utilisant uniquement la méthode Helder
     * @param x Les abscisses des points de support
     * @param y Les ordonnées des points de support
     * @throws IrregularSysLinException Si le système linéaire est irrégulier
     */
    public void identifie(double[] x, double[] y) throws IrregularSysLinException {
        int n = x.length;
        
        // Affichage des points pour débogage
        System.out.println("Points utilisés pour l'identification:");
        for (int i = 0; i < n; i++) {
            System.out.printf("Point %d: (%.4f, %.4f)%n", i+1, x[i], y[i]);
        }
        
        // Construction du système normal A*a = b
        Matrice A = new Matrice(degre + 1, degre + 1);
        Vecteur b = new Vecteur(degre + 1);
        
        // Remplissage de la matrice A et du vecteur b
        for (int i = 0; i <= degre; i++) {
            for (int j = 0; j <= degre; j++) {
                double somme = 0.0;
                for (int k = 0; k < n; k++) {
                    somme += fonctionDeBase(i, x[k]) * fonctionDeBase(j, x[k]);
                }
                A.remplacecoef(i, j, somme);
            }
            
            double somme = 0.0;
            for (int k = 0; k < n; k++) {
                somme += y[k] * fonctionDeBase(i, x[k]);
            }
            b.remplacecoef(i, somme);
        }
        
        // Affichage de la matrice A et du vecteur b pour débogage
        System.out.println("\nMatrice A du système normal:");
        for (int i = 0; i <= degre; i++) {
            for (int j = 0; j <= degre; j++) {
                System.out.printf("%.4f ", A.getCoeff(i, j));
            }
            System.out.println();
        }
        
        System.out.println("\nVecteur b du système normal:");
        for (int i = 0; i <= degre; i++) {
            System.out.printf("%.4f ", b.getCoeff(i));
        }
        System.out.println("\n");
        
        // Résolution du système A*a = b avec Helder uniquement
        Helder solveur = new Helder(A, b);
        coefficients = solveur.resolution();
        System.out.println("Coefficients calculés avec Helder: " + coefficients);
        
        // Vérification des coefficients
        System.out.println("\nVérification des coefficients :");
        Vecteur Ax = Vecteur.produit(A, coefficients);
        System.out.println("A*x : " + Ax);
        System.out.println("b   : " + b);
        
        // Vérification avec quelques points
        System.out.println("\nVérification sur quelques points:");
        for (int i = 0; i < Math.min(5, n); i++) {
            double valeurCalculee = evaluer(x[i]);
            System.out.printf("Point (%.4f, %.4f): p(x) = %.4f, erreur = %.4f%n", 
                x[i], y[i], valeurCalculee, Math.abs(y[i] - valeurCalculee));
        }
    }

    /**
     * Retourne les coefficients du polynôme
     * @return Un vecteur contenant les coefficients
     */
    public Vecteur getCoefficients() {
        return coefficients;
    }
    
    /**
     * Retourne le degré du polynôme
     * @return Le degré du polynôme
     */
    public int getDegre() {
        return degre;
    }

    /**
     * Représentation textuelle du polynôme
     */
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("p(x) = ");
        
        for (int i = 0; i <= degre; i++) {
            double coeff = coefficients.getCoeff(i);
            if (i > 0) {
                sb.append(coeff >= 0 ? " + " : " - ");
                sb.append(Math.abs(coeff));
            } else {
                sb.append(coeff);
            }
            
            if (i > 0) {
                sb.append("x");
                if (i > 1) {
                    sb.append("^").append(i);
                }
            }
        }
        
        return sb.toString();
    }
}